/******************************************************************************
 * $Source: /export/CVS/java/bb_util/src/de/bb/util/ByteRef.java,v $
 * $Revision: 1.30 $
 * $Date: 2008/03/15 18:01:05 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 *
 * A String like class for direct byte manipulation.
 * Since java.nio this might look outdated, but there are still devices
 * where java.nio does not exist
 *
 * Copyright (c) by Stefan Bebbo Franke 1999-2008.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *****************************************************************************/

package de.bb.util;

import java.io.*;

/**
 * <p>A String like class to manipulate byte arrays.
 * The <b>most important</b> differences to Strings:
 * <ul><li><code>ByteRef</code> <b>share their buffer</b></li>
 * <li>the current object <b>can be modified</b> by its functions</li></ul>
 * So the user of <code>ByteRef</code> <b>must</b> handle them correctly. The result is a big speed gain in data parsing and handling.
 * </p><p>
 * For not yet or bad documented functions search the comment in the String class (or guess).
 * </p>
 *
 * @author Stefan Franke
 */
public final class ByteRef implements Cloneable
{
  private final static byte [] nul = new byte[0];
  private byte [] data;
  private int begin;
  private int end;

  /**
   * Initializes a newly created <code>ByteRef</code> object so that it
   * represents an empty byte sequence.
   */
  public ByteRef()
  {
    data = nul;
    begin = end = 0;
  }
  /**
   * Initializes a newly created <code>ByteRef</code> object so that it
   * represents the same sequence of bytes as the argument; in other
   * words, the newly created byte array is a copy of the argument string.getBytes().
   *
   * @param   s   a <code>String</code>.
   */
  public ByteRef(String s)
  {
    if (s != null)
    {
      int len = s.length();
      char ch[] = new char[len];
      s.getChars(0, len, ch, 0);
      byte b[] = data = new byte[len];
      for (int i = 0; i < len; ++i) {
        b[i] = (byte)ch[i];
      }
      begin = 0;
      end = len;
    } else {
      data = nul;
      begin = end = 0;
    }
  }
  /**
   * Initializes a newly created <code>ByteRef</code> object so that it
   * represents the same sequence of bytes as the argument; in other
   * words, the newly created byte array is a copy of the argument.
   *
   * @param   b   a <code>byte array</code>.
   */
  public ByteRef(byte [] b)
  {
    this(b, 0, b.length);
  }
  /**
   * Initializes a newly created <code>ByteRef</code> object so that it
   * represents the same sequence of bytes as the argument; in other
   * words, the newly created byte array is a copy of the specified part of the argument.
   *
   * @param   b   a <code>byte array</code>.
   * @param _begin begin in the byte array
   * @param _end   end in the byte array
   */
  public ByteRef(byte [] b, int _begin, int _end)
  {
    if (_begin < 0)
      _begin = 0;
    else
    if (_begin > b.length)
      _begin = b.length;
      
    if (_end > b.length)
      _end = b.length;
      
    data = b;
    begin = _begin;
    end = _end > _begin ? _end : _begin;
  }
  /**
   * Assigns the parameters physical byte array to this object.
   * @param o a <code>ByteRef</code>
   */
  public void assign(ByteRef o)
  {
    data = o.data;
    begin = o.begin;
    end = o.end;
  }
  /**
   * Assigns the parameters physical byte array to this object.
   * @param b a <code>byte array</code>
   * @param _begin begin in the byte array
   * @param _end   end in the byte array
   */
  public final void assign(byte [] b, int _begin, int _end)
  {
    if (_begin < 0)
      _begin = 0;
    else
    if (_begin > b.length)
      _begin = b.length;
      
    if (_end > b.length)
      _end = b.length;
      
    data = b;
    begin = _begin;
    end = _end > _begin ? _end : _begin;
  }
  /**
   * Extracts the next available line from this object and returns that line as a new <code>BytRef</code> object.
   * Unix and DOS EOLs are recognized correctly. MAC EOLs will fail at end of buffer!
   * @return a new allocated <code>ByteRef</code> object with the next line, or <code>null</code> if no EOL was found.
   */
  public final ByteRef nextLine()
  {
    byte data[] = this.data;
    for (int i = begin; i < end; ++i)
    {
      byte b = data[i];
      // DOS mode
      if (b == 0xd)
      {
        ByteRef r = new ByteRef(data, begin, i);
        ++i;
        if (i == end || data[i] != 0xa) // MAC does it without 0xa
          begin = i;
        else
          begin = i+1;
        return r;
      }
      // unix mode
      if (b == 0xa)
      {
        ByteRef r = new ByteRef(data, begin, i);
        begin = i+1;
        return r;
      }
    }
    return null;
  }
  /**
   * Extracts the next available word from this object and returns that word as a new <code>BytRef</code> object.
   * All bytes < 0x20 are treated as white spaces.
   * @return a new allocated <code>ByteRef</code> object with the next word, or <code>null</code> at end of buffer.
   */
  public final ByteRef nextWord()
  {
    if (begin == end)
      return null;
    int i = begin;
    for (; i < end; ++i)
    {
      if (data[i] <= 32)
        break;
    }
    ByteRef r = new ByteRef(data, begin, i);
    while (i < end && data[i] <= 32)
      ++i;
    begin = i;
    return r;
  }

  /**
   * Extracts the content from this object up to the delimiter and returns that as a new <code>BytRef</code> object.
   * The current object is modified and contains the remaining content behind the delimiter.
   * @param delim the used delimiter
   * @return a new allocated <code>ByteRef</code> object with the next word, or <code>null</code> at end of buffer.
   */
  public final ByteRef nextWord(int delim)
  {
    if (begin == end)
      return null;
    int i = begin;
    for (; i < end; ++i)
    {
      if (data[i] == delim)
        break;
    }
    ByteRef r = new ByteRef(data, begin, i);
    while (i < end && data[i] == delim)
      ++i;
    begin = i;
    return r;
  }

  /**
   * Extracts the content from this object up to the delimiter and returns that as a new <code>BytRef</code> object.
   * The current object is modified and contains the remaining content behind the delimiter.
   * @param delim the used delimiter
   * @return a new <code>String</code> object with the next word, or <code>null</code> at end of buffer.
   */
  public final String nextWordAsString(int delim)
  {
    if (begin == end)
      return null;
    int i = begin;
    for (; i < end; ++i)
    {
      if (data[i] == delim)
        break;
    }
    int len = i - begin;
    char ch[] = new char[len];
    for (int j = 0; j < len; ++j) {
      ch[j] = (char)(0xff & data[begin + j]);
    }
    String r = new String(ch);
    while (i < end && data[i] == delim)
      ++i;
    begin = i;
    return r;
  }
  
  /**
   * Return the length of this<code>ByteRef</code> object.
   * @return the length of this<code>ByteRef</code> object
   */
  public final int length()
  {
    return end - begin;
  }
  /**
   * Copy this data into the specified <code>byte array</code>.
   * @param b a <code>byte array</code>
   */
  public final void copy(byte b[])
  {
    System.arraycopy(data, begin, b, 0, end - begin);
  }
  /**
   * Copy a part of this data into the specified <code>byte array</code> at specified offset.<br>
   * If there are not enough bytes to copy, only the available bytes are copied.
   * @param b a <code>byte array</code>
   * @param off the offset into b
   * @param l the copied length
   * @return an int representing the real count of copied bytes
   */
  public final int copy(byte b[], int off, int l)
  {
    if (l > end - begin)
      l = end - begin;
    System.arraycopy(data, begin, b, off, l);
    return l;
  }
  /**
   * Append the specified <code>ByteRef</code> to this <code>ByteRef</code>.
   * A new <code>ByteRef</code> is allocated and returned, to store the concatenation of both <code>ByteRef</code>s.
   * @param o the appended <code>ByteRef</code>
   * @return a new <code>ByteRef</code> object with the concatenation of both <code>ByteRef</code>s.
   */
  public final ByteRef append(ByteRef o)
  {
    byte b[] = new byte[end - begin + o.end - o.begin];
    System.arraycopy(data, begin, b, 0, end - begin);
    System.arraycopy(o.data, o.begin, b, end - begin, o.end - o.begin);
    return new ByteRef(b);
  }
  /**
   * Append the bytes of the <code>String</code> to this <code>ByteRef</code>.
   * A new <code>ByteRef</code> is allocated and returned, to store the concatenation.
   * @param s the appended <code>String</code>
   * @return a new <code>ByteRef</code> object with the concatenation.
   */
  public final ByteRef append(String s)
  {
    return append(new ByteRef(s));
  }
  /**
   * Returns the <code>ByteRef</code>s data as a <code>String</code>.
   * @return the <code>ByteRef</code>s data as a <code>String</code>.
   */
  public final String toString()
  {
    int len = end - begin;
    char chs[] = new char[len];
    for (int i = 0; i < len; ++i)
      chs[i] = (char)(0xff&data[begin + i]);
    return new String(chs);
  }
  
  /**
   * Returns the <code>ByteRef</code>s data as a <code>String</code> starting from specified offset.
   * @param offset starting from offset
   * @return the <code>ByteRef</code>s data as a <code>String</code>.
   */
  public final String toString(int offset)
  {
    int len = end - begin - offset;
    if (len <= 0)
      return "";
    char chs[] = new char[len];
    for (int i = 0; i < len; ++i)
      chs[i] = (char)(0xff&data[begin + i + offset]);
    return new String(chs);
  }
  
  /**
   * Returns the <code>ByteRef</code>s data as a new byte array.
   * @return the <code>ByteRef</code>s data as a new byte array.
   */
  public final byte [] toByteArray()
  {
    byte b[] = new byte[end-begin];
    System.arraycopy(data, begin, b, 0, b.length);
    return b;
  }
  /**
   * Converts lower case letters into lower case letters.
   * <b>Convertion is done in place!</b>
   * @return this
   */
  public final ByteRef toUpperCase()
  {
    for (int i = begin; i < end; ++i)
    {
      byte b = data[i];
      if (b >= 'a' && b <= 'z')
        data[i] = (byte)(b - 0x20);
    }
    return this;
  }
  /**
   * Converts upper case letters into lower case letters.
   * <b>Convertion is done in place!</b>
   * @return this
   */
  public final ByteRef toLowerCase()
  {
    for (int i = begin; i < end; ++i)
    {
      byte b = data[i];
      if (b >= 'A' && b <= 'Z')
        data[i] = (byte)(b + 0x20);
    }
    return this;
  }
  /**
   * Returns the last index of the specified byte. Search starts et end of data.
   * Is same as br.lastIndexOf(z, br.length()).
   * @param z the searched byte
   * @return the last index of the specified byte, or -1 when not found.
   */
  public final int lastIndexOf(int z)
  {
    return lastIndexOf(z, end - begin);
  }
  /**
   * Returns the last index of the specified byte.
   * @param z the searched byte
   * @param off the offset (from begin) where search starts
   * @return the last index of the specified byte, or -1 when not found.
   */
  public final int lastIndexOf(int z, int off)
  {
    byte b = (byte)z;
    for (int i = begin + off - 1; i >= begin; --i)
      if (data[i] == b)
        return i - begin;
    return -1;
  }
  /**
   * Returns the first index of the specified byte. Search starts et begin of data.
   * Is same as br.indexOf(z, 0).
   * @param z the searched byte
   * @return the first index of the specified byte, or -1 when not found.
   */
  public final int indexOf(int z)
  {
    return indexOf(z, 0);
  }
  /**
   * Returns the first index of the specified byte. Search starts et begin of data.
   * Is same as br.indexOf(z, 0).
   * @param z the searched byte
   * @param off the offset (from begin) where search starts
   * @return the first index of the specified byte, or -1 when not found.
   */
  public final int indexOf(int z, int off)
  {
    byte b = (byte)z;
    for (int i = begin + off; i < end; ++i)
      if (data[i] == b)
        return i - begin;
    return -1;
  }
  /**
   * Returns the first index of the specified String's bytes. Search starts et begin of data.
   * Is same as br.indexOf(s, 0).
   * @param s the searched String
   * @return the first index of the specified byte, or -1 when not found.
   */
  public final int indexOf(String s)
  {
    return indexOf(s, 0);
  }
  /**
   * Returns the first index of the specified String's bytes. Search starts et begin of data.
   * Is same as br.indexOf(s, 0).
   * @param s the searched String
   * @param off the offset (from begin) where search starts
   * @return the first index of the specified byte, or -1 when not found.
   */
  public final int indexOf(String s, int off)
  {
    byte b[] = s.getBytes();
    for (int i = begin + off; i <= end - b.length; ++i)
    {
      int j = 0;
      for (; j < b.length; ++j)
      {
        if (data[i+j] != b[j])
          break;
      }
      if (j == b.length)
        return i - begin;
    }
    return -1;
  }
  /**
   * Returns the first index of the specified ByteRef's bytes. Search starts et begin of data.
   * @param s the searched ByteRef
   * @param off the offset (from begin) where search starts
   * @return the first index of the specified byte, or -1 when not found.
   */
  public final int indexOf(ByteRef s, int off)
  {
    byte b[] = s.data;
    int sLen = s.length();
    for (int i = begin + off; i <= end - sLen; ++i)
    {
      int j = 0;
      for (; j < sLen; ++j)
      {
        if (data[i+j] != b[j + s.begin])
          break;
      }
      if (j == sLen)
        return i - begin;
    }
    return -1;
  }
  /**
   * Returns the first index of the specified ByteRef's bytes. Search starts et begin of data.
   * Is same as br.indexOf(s, 0).
   * @param s the searched ByteRef
   * @return the first index of the specified byte, or -1 when not found.
   */
  public final int indexOf(ByteRef s) {
    return indexOf(s, 0);
  }

  /**
   * Returns a new BytRef object using only the specified part of this <code>ByteRef</code>.
   * @param a begin of used bytes
   * @param z end of used bytes
   * @return a new BytRef object using only the specified part of this <code>ByteRef</code>.
   */
  public final ByteRef substring(int a, int z)
  {
    return new ByteRef(data, begin + a, begin + z);
  }
  /**
   * Returns a new BytRef object using only the specified part of this <code>ByteRef</code>, up to end of this <code>ByteRef</code>.
   * @param a begin of used bytes
   * @return a new BytRef object using only the specified part of this <code>ByteRef</code>.
   */
  public final ByteRef substring(int a)
  {
    return new ByteRef(data, begin + a, end);
  }
  /**
   * Returns the byte at the specified position.
   * @param idx index of the byte
   * @return  the byte at the specified position.
   */
  public final int charAt(int idx)
  {
    if (idx < 0 || begin + idx >= end)
      return -1;
    return data[begin + idx] & 0xff;
  }
  /**
   * Compute a hashCode for use in Hashtable.
   * @return a hashcode.
   */
  public final int hashCode()
  {
    int hc = end - begin;
    for (int i = begin; i < end; ++i)
      hc ^= (hc<<4)^data[i];
    return hc;
  }
  /**
   * Compares this objects data to another object.
   * @param _o  an Object
   * @return true, when other object is also a BytRef and it contains same data.
   */
  public final boolean equals(Object _o)
  {
    if (! (_o instanceof ByteRef))
    {
      if (_o instanceof String)
        return equals(new ByteRef((String)_o));
      return false;
    }
    ByteRef o = (ByteRef)_o;
    if (o.end - o.begin != end - begin)
      return false;
    for (int i = begin, j = o.begin; i < end; ++i, ++j)
      if (data[i] != o.data[j])
        return false;
    return true;
  }
  /**
   * Compares this objects data to another object ignoring upper and lower case.
   * @param _o  an Object
   * @return true, when other object is also a BytRef and it contains same data.
   */
  public final boolean equalsIgnoreCase(Object _o)
  {
    if (! (_o instanceof ByteRef))
    {
      if (_o instanceof String)
        return equalsIgnoreCase(new ByteRef((String)_o));
      return false;
    }
    ByteRef o = (ByteRef)_o;
    if (o.end - o.begin != end - begin)
      return false;
    for (int i = begin, j = o.begin; i < end; ++i, ++j)
    {
      byte a = data[i];
      if (a >= 'a' && a <= 'z')
        a = (byte)(a - 0x20);
      byte b = o.data[j];
      if (b >= 'a' && b <= 'z')
        b = (byte)(b - 0x20);

      if (a != b)
        return false;
    }
    return true;
  }
  /**
   * Compares this <code>ByteRef</code>'s data to another <code>ByteRef</code>.
   * @param o  a ByteRef object
   * @return 0, when other <code>ByteRef</code> contains same data, -1 when this is less and 1 when this is greater then the other <code>ByteRef</code>.
   */
  public final int compareTo(ByteRef o)
  {
    for (int i = begin, j = o.begin; i < end && j < o.end; ++i, ++j)
    {
      int diff = data[i] - o.data[j];
      if (diff != 0)
        return diff;
    }
    return begin - end - o.begin + o.end;
  }
  /**
   * Tests whether this BytRef ends with the specified <code>ByteRef</code>.
   * @param o the other <code>ByteRef</code> where the end is tested against
   * @return true if this BytRef ends with the specified <code>ByteRef</code>
   */
  public final boolean endsWith(ByteRef o)
  {
    int i = end + o.begin - o.end, j = o.begin;
    if (i < begin)
      return false;
    for (; i < end; ++i, ++j)
    {
      int diff = data[i] - o.data[j];
      if (diff != 0)
        return false;
    }
    return true;  
  }
  /**
   * Tests whether this BytRef ends with the specified String.
   * @param s the other String where the end is tested against
   * @return true if this BytRef ends with the specified <code>String</code>
   */
  public final boolean endsWith(String s)
  {
    return endsWith(new ByteRef(s));
  }

  /**
   * Tests whether this BytRef starts with the specified <code>ByteRef</code>.
   * @param o the other <code>ByteRef</code> where the end is tested against
   * @return true if this BytRef starts with the specified <code>ByteRef</code>
   */
  public final boolean startsWith(ByteRef o)
  {
    int i = begin, j = o.begin;
    for (; i < end && j < o.end; ++i, ++j)
    {
      int diff = data[i] - o.data[j];
      if (diff != 0)
        return false;
    }
    return j == o.end;  
//    return substring(0, o.length()).compareTo(o) == 0;
  }
  /**
   * Tests whether this BytRef starts with the specified String.
   * @param s the other String where the end is tested against
   * @return true if this BytRef starts with the specified <code>String</code>
   */
  public final boolean startsWith(String s)
  {
    return startsWith(new ByteRef(s));
  }


  /**
   * Write the bytes to the specified output stream.
   * @param os a output stream
   * @throws IOException on io errors
   */
  public final void writeTo(OutputStream os) throws IOException
  {
    os.write(data, begin, end - begin);
  }
  /**
   * Remove all occurencies of the specified charater at begin and end.
   * @param ch a character value
   * @return a new <code>ByteRef</code> with the trimed data.
   */
  public final ByteRef trim(int ch)
  {
    int b = begin, e = end;
    while (b < e && (data[b]&0xff) == ch)
      ++b;
    while (b < e && (data[e-1]&0xff) == ch)
      --e;
    return new ByteRef(data, b, e);
  }
  /**
   * Remove all white spaces at begin and end.
   * @return a new <code>ByteRef</code> with the trimed data.
   */
  public final ByteRef trim()
  {
    int b = begin, e = end;
    while (b < e && (data[b]&0xff) <= 32)
      ++b;
    while (b < e && (data[e-1]&0xff) <= 32)
      --e;
    return new ByteRef(data, b, e);
  }
  /**
   * Interpret the bytes as characters and convert them into an int value.
   * @return an int value
   */
  public final int toInteger()
  {
    int z = 0;
    for (int i = begin; i < end; ++i)
    {
      int c = data[i]&0xff;
      if (c < '0' || c > '9')
        break;
      z *= 10;
      z += c - '0';
    }
    return z;
  }
    
  /**
   * Update this <code>ByteRef</code> with the availabledata of the specified input stream.
   * <b>current object's data is also updated!</b>.
   * @param is an input stream
   * @return this on success with old and appended new data, or null on failure.
   */
  public final ByteRef update(InputStream is)
  {
    try 
    {
      // try read a single byte
      int firstByte = is.read();
      if (firstByte == -1)
        return null;
      
      // old length + already read byte
      int baseLen = length() + 1;
      
      // how many bytes are available?
      int avail = is.available();
      // limit readahaed      
      if (avail > 0x4000) {
        avail = 0x4000;
      }
            
      byte [] r = new byte[baseLen + avail];
      // copy existing data to buffer
      copy(r);
      // add single read byte
      r[baseLen-1] = (byte)firstByte;
      
      // how many bytes are really read?
      int count = 0;
      if (avail > 0) {
        count = is.read(r, baseLen, avail);
      }
      data = r;
      begin = 0;
      end = baseLen + count;
//      assign(r, 0, baseLen + count);
      return this;
    } catch (Exception e)
    {//
    }
    return null;
  }
  /**
   * Extracts the next line from the specified <code>ByteRef</code> br.
   * If there is no complete line in br, the <code>ByteRef</code> is updated using the specified input stream.
   * This is done until a complete line is read, or br.update(is) fails.
   * @param br the <code>ByteRef</code> used as input buffer
   * @param is a input stream
   * @return a new <code>ByteRef</code> containing the next line, or null on EOS.
   */
  public final static ByteRef readLine(ByteRef br, InputStream is)
  {
    // get request line
    while(br != null)
    {
      // get the next line
      // if (DEBUG) System.out.println("C: " + br.toString());
      ByteRef req = br.nextLine();
      if (req != null)
        return req;
      br = br.update(is);
    }
    return null;
  }

  public final Object clone()
  {
    return new ByteRef(this.toByteArray());
  }
  /**
   * Returns a ByteRef starting from 0 to stop, 
   * and removes this part from the current ByteRef.
   * This ByteRef contains only the content from stop to end.
   * 
   * This is a self modifying method!!
   * @param stop end position 
   * @return a ByteRef as substring(0, stop) but modifies this object.
   */
  public final ByteRef splitLeft(int stop) {
    return splitLeft(0, stop);
  }

  /**
   * Returns a ByteRef starting from start to stop, 
   * and removes this part from the current ByteRef.
   * This ByteRef contains only the content from stop to end.
   * 
   * This is a self modifying method!!
   * @param start start position for returned ByteRef
   * @param stop stop position for returned ByteRef
   * @return a ByteRef as substring(start, stop) but modifies this object.
   */
  public final ByteRef splitLeft(int start, int stop) {
    if (stop < 0)
      stop = 0; 
    ByteRef r = substring(start, stop);
    begin += stop;
    if (begin > end)
      begin = end;
    return r;
  }
  /**
   * Remove the first byte and return it as int >= 0 or return -1;
   * This is a self modification.
   * @return the first byte and return it as int >= 0 or return -1;
   */
  public final int removeLeft()
  {
    if (begin == end)
      return -1;
    return 0xff & data[begin++];
  }
  /**
   * Move the left offset by count bytes.
   * This is a self modification.
   * @param count the count of bytes to remove from left side.
   */
  public final void adjustLeft(int count)
  {
    begin += count;
    if (begin > end)
      begin = end;
  }
}

/******************************************************************************
 * $Log: ByteRef.java,v $
 * Revision 1.30  2008/03/15 18:01:05  bebbo
 * @R Changed the license: From now on GPL 3 applies.
 *
 * Revision 1.29  2008/03/13 20:51:02  bebbo
 * @R made class and all methods final
 *
 * Revision 1.28  2006/10/04 08:15:30  bebbo
 * @I better variable names
 *
 * Revision 1.27  2006/05/09 08:54:24  bebbo
 * @I removed unused variables
 *
 * Revision 1.26  2006/03/17 11:37:37  bebbo
 * @B changed toString behaviour: high byte is now 0
 *
 * Revision 1.25  2006/02/02 07:51:20  bebbo
 * @I cleanup
 *
 * Revision 1.24  2004/04/14 13:57:51  bebbo
 * @B fixed JavaDoc issues
 *
 * Revision 1.23  2004/03/14 18:01:13  bebbo
 * @I implementet startsWith() and endsWith() without using substring()
 * @N added splitLeft() method which MODIFIES current object and returns left part
 *
 * Revision 1.22  2002/09/30 07:34:03  bebbo
 * @B update() could be very slow - changed readahead buffer size
 *
 * Revision 1.21  2002/05/16 15:25:29  franke
 * @I modified update()
 *
 * Revision 1.20  2002/01/22 09:38:18  bebbo
 * @B fixed equalsIgnoreCase()
 *
 * Revision 1.19  2002/01/22 09:04:04  bebbo
 * @C javadoc
 *
 * Revision 1.18  2002/01/19 15:47:21  franke
 * @N added equalsIgnoreCase(), clone()
 * @I added minimal read in update()
 *
 * Revision 1.17  2002/01/16 15:32:11  franke
 * @N added more support functions
 *
 * Revision 1.16  2001/12/10 11:56:07  bebbo
 * @C completed comments!
 *
 * Revision 1.15  2001/09/15 08:55:41  bebbo
 * @C added many comments
 *
 * Revision 1.14  2001/08/24 08:23:44  bebbo
 * @R renamed functions - same names as in String class
 *
 * Revision 1.13  2001/08/03 13:48:57  bebbo
 * @B fixed EOL doubling at buffer end
 *
 * Revision 1.12  2001/05/14 21:40:23  bebbo
 * @B removed an optimization which caused bugs
 *
 * Revision 1.11  2001/05/07 16:20:54  bebbo
 * @I improved (really?) method update(): try to reuse existing buffer
 *
 * Revision 1.10  2001/04/16 13:36:22  bebbo
 * @N added nextWord()
 * @N added compareTo(ByteRef)
 * @N added append(String )
 *
 * Revision 1.9  2001/03/28 13:28:46  bebbo
 * @B final fix in update(InputStream is)
 *
 * Revision 1.8  2001/03/28 10:27:21  bebbo
 * @B fixed function update() - did not wait for data properly
 *
 * Revision 1.7  2001/03/11 20:42:12  bebbo
 * @B fixed indexOf
 *
 * Revision 1.6  2001/03/09 19:50:13  bebbo
 * @B assign now performs more checks
 *
 * Revision 1.5  2001/02/25 17:06:17  bebbo
 * @N added static mehtod readLine
 *
 * Revision 1.4  2001/01/15 16:32:54  bebbo
 * @I rewrote nextLine()
 * @N added charAt()
 *
 * Revision 1.3  2001/01/01 16:53:06  bebbo
 * @R added CT for byte[]
 * @R added toByteArray()
 *
 * Revision 1.2  2000/12/29 17:47:03  bebbo
 * @N new functions
 *
 * Revision 1.1  2000/11/10 18:12:26  bebbo
 * @N new
 *
 *****************************************************************************/