/******************************************************************************
 * $Source: /export/CVS/java/bb_util/src/de/bb/util/FileBrowser.java,v $
 * $Revision: 1.6 $
 * $Date: 2008/03/15 18:01:05 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 *
 * Recurse over directories using a mask and invoke callbacks for files and directories.
 *
 * Copyright (c) by Stefan Bebbo Franke 1999-2008.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *****************************************************************************/

package de.bb.util;

import java.io.File;
import java.util.LinkedList;

/**
 * This is a base class to iterate over files and directories, which can be configured with a pattern.
 * @author bebbo
 * @version $Revision: 1.6 $
 */
public abstract class FileBrowser
{
  /** bytes of a defined mask. */
  private byte[] maskBytes;
  private String baseDir;
  private boolean recurse, used;
  private LinkedList list = new LinkedList();
  /**
   * ct.
   * @param mask a filter which is applied to the files.
   */
  public FileBrowser(String mask)
  {
    maskBytes = mask.getBytes();
  }

  /**
   *  Handle a file, matching the specified masks.
   * @param path path to file.
   * @param file the file name without path.
   */
  protected abstract void handleFile(String path, String file);
  /**
   * Handle a directory found in that path.
   * @param path path to file.
   * @param dir the dirctories name without path.
   */
  protected abstract void handleDir(String path, String dir);
  /**
   * Compare the file name with the specified pattern.
   * The pattern is held in the global variable maskBytes
   * @param name the checked file name
   * @return true if the file name matches the pattern, false either.
   */
  protected boolean accept(String name)
  {
    byte[] n = name.getBytes();
    // now compare this with the maskBytes
    int i = 0, j = 0;
    boolean star = false, sstar = false;
    int is = 0, js = 0;
    try
    {
      for (; i < maskBytes.length;)
      {
        if (maskBytes[i] == '*')
        {
          ++i;
          star = true;          
          if (i < maskBytes.length && maskBytes[i] == '*')
          {
            sstar = true;
            ++i;
            is = i;
            js = j;
          } else {
            sstar = false;
          }
          continue;
        }

        if (maskBytes[i] == n[j] || maskBytes[i] == '?')
        {
          ++i;
          ++j;
          star = false;
          continue;
        }
        if (!star) {
          if (sstar) {
            j = ++js;
            if (j == n.length)
              return false;

            i = is;
            star = true;
            continue;
          }
          break;
        }
        ++j;
      }
    } catch (Exception e)
    {
      //
    }
    return i == maskBytes.length && (star || j == n.length);
  }
  /**
   * Recurse over directories and invoke the file action for each File.
   * @param path path relative to baseDir
   */
  private boolean recurseDirs(String path)
  {
    boolean empty = used;
    File dir = new File(baseDir, path);

    String[] files = dir.list();
    for (int i = 0; files != null && i < files.length; ++i)
    {
      File f = new File(dir, files[i]);
      if (f.isDirectory())
      {
        if (recurse)
        {

          list.add(new String[] { path, files[i] });
          String s = path + "/" + files[i];
          boolean subempty = recurseDirs(s);

          empty &= subempty;
        }
      } else
      {
        if (accept(path + "/" + files[i]))
        {
          if (empty)
          {
            while (list.size() > 0)
            {
              String[] ss = (String[]) list.removeFirst();
              handleDir(ss[0], ss[1]);
            }
            empty = false;
          }
          handleFile(path, files[i]);
        }
      }
    }
    if (empty && list.size() > 0)
    {
      list.removeLast();
    }
    return empty;
  }
  /**
   * Recurse over directories and invoke the file action for each File.
   * @param baseDir the dir to start
   * @param recurse a flag whether subdirectories are scanned too
   * @param suppressEmtpyDirs whether empty dirs are emitted or not
   */
  public final synchronized void scan(
    String baseDir,
    boolean recurse,
    boolean suppressEmtpyDirs)
  {
    if (!baseDir.endsWith("/"))
      baseDir += "/";
    this.baseDir = baseDir;
    this.recurse = recurse;
    this.used = suppressEmtpyDirs;
    recurseDirs("");
  }
  /**
   * Returns the baseDir.
   * @return String
   */
  public String getBaseDir()
  {
    return baseDir;
  }

}

/*
 * Log: $Log: FileBrowser.java,v $
 * Log: Revision 1.6  2008/03/15 18:01:05  bebbo
 * Log: @R Changed the license: From now on GPL 3 applies.
 * Log:
 * Log: Revision 1.5  2007/01/18 22:02:34  bebbo
 * Log: @N added support/different behaviour for ** and *
 * Log:
 * Log: Revision 1.4  2006/02/02 07:49:02  bebbo
 * Log: @I cleanup
 * Log:
 * Log: Revision 1.3  2003/11/26 10:02:28  bebbo
 * Log: @B fix for missing "/" at end of path
 * Log:
 * Log: Revision 1.2  2003/04/02 12:15:36  bebbo
 * Log: @C resolved API Doc errors
 * Log:
 * Log: Revision 1.1  2002/11/19 12:58:37  bebbo
 * Log: @N new
 * Log:
 */