/******************************************************************************
 * $Source: /export/CVS/java/bb_util/src/de/bb/util/LRUCache.java,v $
 * $Revision: 1.14 $
 * $Date: 2008/03/15 18:01:05 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 *
 * An efficient LRU cache implementation  
 *
 * Copyright (c) by Stefan Bebbo Franke 1999-2008.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *****************************************************************************/

package de.bb.util;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.HashSet;

/**
 * A 2-level LRU cache implementation.
 * - Both levels are implemented as global LRU lists. All cache instances are sharing the global LRU lists.
 * - The specified percentage defines the count of entries in level 2 to total count of entries.
 * - Each new entry is placed into the level 1.
 * - If a lookup succeeds and the entry is in level 1, it gets promoted to level 2.
 * - If level 2 exceeds the allowed count it gets downgraded to level 1.
 * - All entries in level 1 are soft references, and can be cleared by the gc.
 *   Since this reduces the total and further usages will establish the specified percentage.
 * - All entries in level 2 are hard references, so the often used entries are rarely discarded.
 * - There is no configuration but the ratio for level 2.
 * 
 * So the contract is: If you put something into the cache, it may be there if you look after it - or not.
 * This is a cache and not a HashMap!
 * 
 * @author Bebbo
 * @version $Revision: 1.14 $
 */
public class LRUCache {

  static ReferenceQueue RQ = new ReferenceQueue();
  static RQThread RQTHREAD;
  private static Node ROOT_HARD = new Node();
  private static Node ROOT_SOFT = new Node();
  
    
  private HashMap hashMap = new HashMap();  
  private int countHard;
  private int ratio;
  // statistic
  private int hit;
  private int get;

  /**
   * Create a new LRUCache with default ratio of 10% fixed elements (level 2).
   * Use a dynamic size. 
   */
  public LRUCache() {
    this(0.1);
  }
  
  /**
   * Creates a new LRUCache with the specified ratio in percent (level 2).
   * Use a dynamic size.
   * @param ratio a percent value 0.01 &lt;= ratio &lt;= 0.99
   */
  public LRUCache(double ratio) {
    if (ratio < 0.001) ratio = 0.001;
    else if (ratio > 0.999) ratio = 0.999;
    
    this.ratio = (int)(ratio * 1000);
  }

  /**
   * Return the current count of hard and soft references.
   * @return the current count of hard and soft references.
   */
  public int size() {
    return hashMap.size();
  }
  
  private final boolean hardRefsFull() {
    int countSoft = hashMap.size() - countHard;
    // check whether to convert hard refs to soft refs
    return countHard * 1000 > countSoft * ratio;
  }
  
  /**
   * Put an object into this cache.
   * There is no warranty how long an object is held in this cache!
   * @param key a key to identify the object
   * @param o the sored object
   */
  public synchronized void put(Object key, Object o) {
    // check whether to convert hard refs to soft refs
    if (hardRefsFull()) {
      Node node = ROOT_HARD.prev;
      node.remove();
      node.hardValueRef = null;
      ROOT_SOFT.addHead(node);
      --countHard;
    }
    
    // store new object 
    Node node = new Node(key, o, RQ, this);
//    Node node = new Node(key, o, null, this);
    ROOT_SOFT.addHead(node);
    hashMap.put(key, node);
  }
  
  /**
   * retrieves the object by key, if it is still in the cache.
   * @param key the key to get the object for
   * @return the object if it is still in cache, null either.
   */
  public synchronized Object get(Object key) {
    ++get;
    // check whether object is in cache
    Node n = (Node) hashMap.get(key);
    if (n == null)
      return null;

    // remove node from List - we do LRU so remove/insert
    n.remove();
    
    // check whether SoftReference still knows the object
    Object o = n.get();
    if (o == null) {
      // remove this key from hashtable
      hashMap.remove(key);    
      return null;
    }
    
    ++hit;
    // retain hard references
    if (n.parent == ROOT_HARD) {
      ROOT_HARD.addHead(n);
      return o;
    }

    if (hardRefsFull()) {
      ROOT_SOFT.addHead(n);
      return o;
    }
    
    // modify counter on transition from soft to hard
    // and store the hardValueRef      
    n.hardValueRef = o;
    ++countHard;
    ROOT_HARD.addHead(n);
    return o;      
  }

  /**
   * Method remove.
   * @param key the key for the object to remove
   * @return returns the removed object.
   */
  public synchronized Object remove(Object key)
  {
    // check whether object is in cache
    Node n = (Node) hashMap.remove(key);
    if (n == null)
      return null;

    if (n.parent == ROOT_HARD) {
      --countHard;
      n.hardValueRef = null;
    }
    
    // remove node
    n.remove();
    return n.get();
  }
  


  /**
   * Clears the cache.
   */
  public synchronized void clear()
  {
    // Java likes: create new objects and let GC do the rest!
    hashMap = new HashMap();
    countHard = 0;
  }

  /**
   * Provides a snapshot of keys used in the cache.
   * The HashSet is snapshot which was valid, but may not be accurate now.
   * There is no warranty that any value is still present in the cache 
   * or that nothing else is in the cache.
   * @return a HashSet containing a snapshot of keys used in the cache.
   */
  public HashSet keySet() {
    HashSet hs = new HashSet();
    hs.addAll(hashMap.keySet());
    return hs;
  }
  
  /**
   * Return some information about the current LRUCache object.
   */
  public String toString()
  {
    String r = "" + (100.*hit/(get+1));
    int dot = r.indexOf('.');
    if (dot >= 0 && dot + 2 < r.length()) r = r.substring(0, dot + 3);
    return "bb_util LRUCache: ratio=" + (ratio/1000.) + "; size=" + hashMap.size() + "(hardRefs=" + countHard + "); gets=" + get + ";hit=" + hit + " => " + r + "%";
  }
  

  /**
   * Helper class for a simple double linked list
   */
  private static class Node extends SoftReference {
    Node next, prev;
    Node parent;
    LRUCache cache;
    Object key;
    Object hardValueRef;

    Node()
    {
      super(null);
      next = prev = this;
    }

    Node(Object key, Object val, ReferenceQueue rq, LRUCache cache) {
      super(val, rq);
      this.key = key;
      this.cache = cache;
    }

    /**
     * Remove this node and mark next as zero.
     */
    void remove() {
      if (parent == null) {
        if (prev != null)
          prev.next = next;
        if (next != null)
          next.prev = prev;      
        next = prev = null;
        return;
      }
      // same code but synchronized
      synchronized (parent) {
        if (prev != null)
          prev.next = next;
        if (next != null)
          next.prev = prev;          
        next = prev = null;
      }      
    }

    /**
     * append the node to this root.
     * @param n the node.
     * @return the node.
     */
    Node addHead(Node n) {
      // synchronized to avoid interferences with remove
      synchronized (this) {
        n.next = next;
        n.prev = this;
        n.parent = this;
        if (next != null)      
          next.prev = n;
        next = n;
      }
      return n;      
    }
  }
  
/**/

  /**
   * Poll the Phantom Queue and perform necessary cleanup: remove from Hashtable
   * is always called from synchronized methods, so no additional synchronized needed. 
   */  
  static void clearSoftReferences()
  {
    // remove invalid softreferences from hashtable
    Node n;
    while ((n = (Node)RQ.poll()) != null)
    {
      n.cache.remove(n.key);
    }
  }

  private static class RQThread extends Thread {
    RQThread() {
      super("bb_util: LRUCache cleaning thread.");
      try {
        // try to run as deamon
        setDaemon(true);
        new F();
      } catch (Throwable t){}
    }
    public void run() {
      for(;;) {
      try {
          synchronized (this) {
            this.wait();
          }
          clearSoftReferences();
        } catch (Throwable t) {        
        }
      }
    }    
  }

  static class F {
    public void finalize() {
      synchronized(RQTHREAD) {
        RQTHREAD.notify();      
      }
      new F();
    }
  }
  
  static {
    RQTHREAD = new RQThread();
    RQTHREAD.start();
  }
  /**/
}
/**
 * Log: $Log: LRUCache.java,v $
 * Log: Revision 1.14  2008/03/15 18:01:05  bebbo
 * Log: @R Changed the license: From now on GPL 3 applies.
 * Log:
 * Log: Revision 1.13  2008/01/17 17:40:26  bebbo
 * Log: @I modified the behaviour, so all instances share the same queue to incireas overall efficiency.
 * Log:
 * Log: Revision 1.12  2007/01/18 22:03:34  bebbo
 * Log: @R redesign of LRUCache: it uses now hard and soft references
 * Log:
 * Log: Revision 1.11  2006/03/17 11:37:52  bebbo
 * Log: @B clear is now synchronized
 * Log:
 * Log: Revision 1.10  2004/07/16 09:37:41  bebbo
 * Log: @R remove now returns the Object, if it exists
 * Log:
 * Log: Revision 1.9  2004/04/14 13:57:51  bebbo
 * Log: @B fixed JavaDoc issues
 * Log:
 * Log: Revision 1.8  2004/04/07 16:45:20  bebbo
 * Log: @R redesigned implementation
 * Log:
 * Log: Revision 1.7  2003/09/30 13:17:19  bebbo
 * Log: @B fixed api doc
 * Log:
 * Log: Revision 1.6  2003/09/01 12:49:49  bebbo
 * Log: @N new public method clear(), to clear the cache.
 * Log:
 * Log: Revision 1.5  2003/07/30 10:13:01  bebbo
 * Log: @R default size is now unlimited
 * Log:
 * Log: Revision 1.4  2003/04/02 12:15:36  bebbo
 * Log: @C resolved API Doc errors
 * Log:
 */
