/******************************************************************************
 * $Source: /export/CVS/java/bb_util/src/de/bb/util/MapBase.java,v $
 * $Revision: 1.14 $
 * $Date: 2008/03/15 18:01:05 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 *
 * Base class for map implementations.  
 *
 * Copyright (c) by Stefan Bebbo Franke 1999-2008.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *****************************************************************************/

package de.bb.util;

import java.io.Serializable;
import java.util.AbstractSet;
import java.util.Comparator;
import java.util.Iterator;
import java.util.SortedSet;

/**
 * This is the base class for SingleMap and MultiMap and contains the common part of both implementations.
 * The implementation uses an estimate AVL sorted tree.
 */
abstract class MapBase extends java.util.AbstractMap implements java.util.SortedMap, Serializable {
  /// object to compare two keys
  private Comparator comp;

  /// the root of all stored items
  MapBase.Leaf root;

  /// count of items in tree
  int count;

  /**
   * helper class for elements 
   */
  static class Leaf implements java.util.Map.Entry, Serializable {
    /**
     * Comment for <code>serialVersionUID</code>
     */
    private static final long serialVersionUID = 3256999977816371506L;

    /// refs to top, left nd right
    MapBase.Leaf top, l, r;

    /// left = -1, equal = 0, right = 1
    int len;

    /// the maintained object
    Object key, value;

    /**
     * creates an empty leaf
     * @param k
     * @param v
     */
    Leaf(Object k, Object v)
    {
      top = l = r = null;
      len = 0;
      key = k;
      value = v;
    }

    /** (non-Javadoc)
     * @see java.util.Map.Entry#getKey()
     */
    public Object getKey()
    {
      return key;
    }

    /** (non-Javadoc)
     * @see java.util.Map.Entry#getValue()
     */
    public Object getValue()
    {
      return value;
    }

    /** (non-Javadoc)
     * @see java.util.Map.Entry#setValue(java.lang.Object)
     */
    public Object setValue(Object o)
    {
      Object l = value;
      value = o;
      return l;
    }

    /**
     * get the next - sorting order is defined by tree
     * @return the next Leaf or null
     */
    MapBase.Leaf next()
    {
      Leaf i = this;
      if (i.r != null)
      {
        for (i = i.r; i.l != null;)
          i = i.l;
        return i;
      }
      Leaf j = i.top;
      for (; j != null && j.r == i;)
      {
        i = j;
        j = i.top;
      }
      return j;
    }

    /**
     * get the previous - sorting order is defined by tree
     * @return the next Leaf or null
     */
    Leaf prev()
    {
      Leaf i = this;
      if (i.l != null)
      {
        for (i = i.l; i.r != null;)
          i = i.r;
        return i;
      }
      Leaf j = i.top;
      for (; j != null && j.l == i;)
      {
        i = j;
        j = i.top;
      }
      return j;
    }
    
    public String toString() {
      return key + "=" + value;
    }
  }

  private static class SubSet extends AbstractSet implements SortedSet {
    private MapBase mb;

    private Object fromKey, toKey;

    SubSet(Object from, Object to, MapBase _mb)
    {
      fromKey = from;
      toKey = to;
      mb = _mb;
    }

    private boolean inRange(Object key)
    {
      if (fromKey != null && mb.compare(fromKey, key) > 0)
        return false;
      if (toKey != null && mb.compare(toKey, key) <= 0)
        return false;
      return true;
    }

    /** (non-Javadoc)
     * @see java.util.SortedSet#comparator()
     */
    public Comparator comparator()
    {
      return mb.comparator();
    }

    /** (non-Javadoc)
     * @see java.util.SortedSet#first()
     */
    public java.lang.Object first()
    {
      if (fromKey == null)
        return mb.firstLeaf();
      MapBase.Leaf l = mb.findNext(fromKey);
      if (l == null || !inRange(l.key))
        return null;
      return l;
    }

    /** (non-Javadoc)
     * @see java.util.SortedSet#headSet(java.lang.Object)
     */
    public java.util.SortedSet headSet(java.lang.Object to)
    {
      if (!inRange(to))
        throw new IllegalArgumentException("key out of range");
      return new SubSet(fromKey, to, mb);
    }

    /** (non-Javadoc)
     * @see java.util.SortedSet#last()
     */
    public java.lang.Object last()
    {
      MapBase.Leaf l;
      if (toKey == null) { 
        l = mb.lastLeaf();
        if (l == null || !inRange(l.key))
          return null;
        return l;
      }
      
      l = mb.findNext(toKey);
      if (l == null) {
        l = mb.lastLeaf();
        if (l == null)        
          return null;
        if (!inRange(l.key))
          return null;
        return l;
      }
      MapBase.Leaf l2 = l.prev();
      if (l2 == null || !inRange(l2.key))
        return null;
      return l2;
    }

    /** (non-Javadoc)
     * @see java.util.SortedSet#subSet(java.lang.Object, java.lang.Object)
     */
    public java.util.SortedSet subSet(java.lang.Object from, java.lang.Object to)
    {
      if (!inRange(to) || !inRange(from))
        throw new IllegalArgumentException("key out of range");
      return new SubSet(from, to, mb);
    }

    /** (non-Javadoc)
     * @see java.util.SortedSet#tailSet(java.lang.Object)
     */
    public java.util.SortedSet tailSet(java.lang.Object from)
    {
      if (!inRange(from))
        throw new IllegalArgumentException("key out of range");
      return new SubSet(from, toKey, mb);
    }

    /** (non-Javadoc)
     * @see java.util.Set#iterator()
     */
    public java.util.Iterator iterator()
    {
      return new Iter(mb, (MapBase.Leaf) first(), (MapBase.Leaf) last());
    }

    /** (non-Javadoc)
     * @see java.util.Set#size()
     */
    public int size()
    {
      if (fromKey == null && toKey == null)
        return mb.size();
      int sz = 0;
      for (Iterator i = iterator(); i.hasNext(); i.next())
        ++sz;
      return sz;
    }

    /** (non-Javadoc)
     * @see java.util.Set#isEmpty()
     */
    public boolean isEmpty()
    {
      return first() == null;
    }

    public String toString()
    {
      StringBuffer sb = new StringBuffer("{");
      for (Iterator i = iterator(); i.hasNext();)
      {
        sb.append(i.next());
        if (i.hasNext())
          sb.append(", ");
      }
      return sb.append("}").toString();
    }

    private static class Iter implements Iterator, Cloneable {
      MapBase mb;

      MapBase.Leaf here, last, end;

      Iter(MapBase m, MapBase.Leaf from, MapBase.Leaf to)
      {
        mb = m;
        here = from;
        last = null;
        end = to;
      }

      /** (non-Javadoc)
       * @see java.util.Iterator#hasNext()
       */
      public boolean hasNext()
      {
        return last != end && here != null;
      }

      /** (non-Javadoc)
       * @see java.util.Iterator#next()
       */
      public Object next()
      {
        last = here;
        here = here.next();
        return last;
      }

      /** (non-Javadoc)
       * @see java.util.Iterator#remove()
       */
      public void remove()
      {
        mb.remove(last.key, last.value);
      }

      /** (non-Javadoc)
       * @see java.lang.Object#clone()
       */
      public Object clone()
      {
        Iter i = new Iter(mb, here, end);
        i.last = last;
        return i;
      }
    }
  }

  /**
   * creates a Map object
   * @param comp the used comparator
   */
  MapBase(Comparator comp)
  {
    this.comp = comp;
    root = null;
    count = 0;
  }

  /**
   * creates a Map object
   */
  MapBase()
  {
    this(null);
  }

  // ============================================================================
  // public functions
  // ============================================================================
  /**
   * Clears this Map so that it contains no objects.
   */
  public void clear()
  {
    root = null;
    count = 0;
    // rest does the garbage collection
  }

  /**
   * Returns the number of components in this Map.
   * @return the number of components in this Map.
   */
  public int size()
  {
    return count;
  }

  /**
   * Tests if this Map has no components.
   * @return true if this Map has no components; false otherwise.
   */
  public final boolean isEmpty()
  {
    return count == 0;
  }

  /**
   * get the sorted first leaf.
   * @return the sorted first leaf.
   */
  MapBase.Leaf firstLeaf()
  {
    if (root == null)
      return null;
    MapBase.Leaf i = root;
    while (i.l != null)
      i = i.l;
    return i;
  }

  /**
   * get the sorted last leaf.
   * @return the sorted last leaf.
   */
  MapBase.Leaf lastLeaf()
  {
    if (root == null)
      return null;
    MapBase.Leaf i = root;
    while (i.r != null)
      i = i.r;
    return i;
  }

  private final void drr(MapBase.Leaf i, MapBase.Leaf j)
  {
    //          System.out.println("double rot right");
    MapBase.Leaf k = j.top;
    if (k == null)
      root = i.r;
    else if (k.l == j)
      k.l = i.r;
    else
      k.r = i.r;
    i.r.top = k;

    k = i.r;

    i.r = k.l;
    if (i.r != null)
      i.r.top = i;
    i.top = k;
    i.len = k.len <= 0 ? 0 : -1;

    j.l = k.r;
    if (j.l != null)
      j.l.top = j;
    j.top = k;
    j.len = k.len >= 0 ? 0 : 1;

    k.l = i;
    k.r = j;
    k.len = 0;
  }

  private final void rr(MapBase.Leaf i, MapBase.Leaf j)
  {
    //        System.out.println("rot right");
    MapBase.Leaf k = j.top;
    if (k == null)
      root = i;
    else if (k.l == j)
      k.l = i;
    else
      k.r = i;

    j.l = i.r;
    if (j.l != null)
      j.l.top = j;
    j.top = i;

    i.r = j;
    i.top = k;
    ++i.len;
    //    if (i.len != 0) System.out.println("*");
    j.len = -i.len;
  }

  private final void drl(MapBase.Leaf i, MapBase.Leaf j)
  {
    //          System.out.println("double rot left");
    MapBase.Leaf k = j.top;
    if (k == null)
      root = i.l;
    else if (k.l == j)
      k.l = i.l;
    else
      k.r = i.l;
    i.l.top = k;
    k = i.l;

    i.l = k.r;
    if (i.l != null)
      i.l.top = i;
    i.top = k;
    i.len = k.len >= 0 ? 0 : 1;

    j.r = k.l;
    if (j.r != null)
      j.r.top = j;
    j.top = k;
    j.len = k.len <= 0 ? 0 : -1;

    k.r = i;
    k.l = j;
    k.len = 0;
  }

  private final void rl(MapBase.Leaf i, MapBase.Leaf j)
  {
    //        System.out.println("rot left");
    MapBase.Leaf k = j.top;
    if (k == null)
      root = i;
    else if (k.l == j)
      k.l = i;
    else
      k.r = i;

    j.r = i.l;
    if (j.r != null)
      j.r.top = j;
    j.top = i;

    i.l = j;
    i.top = k;
    --i.len;
    //    if (i.len != 0) System.out.println("*");
    j.len = -i.len;
  }

  /**
   * Rebalance the tree on add.
   * @param i
   */
  protected final void fixAdd(MapBase.Leaf i)
  {
    for (MapBase.Leaf j = i.top; j != null; i = j, j = i.top)
    {
      if (j.l == i)
        --j.len;
      else
        ++j.len;

      if (j.len == 0)
      {
        return;
      }
      if (j.len == -2)
      {
        if (i.len > 0)
        {
          drr(i, j);
          return;
        }
        rr(i, j);
        return;
      }
      if (j.len == 2)
      {
        if (i.len < 0)
        {
          drl(i, j);
          return;
        }
        rl(i, j);
        return;
      }
    }
  }

  /**
   * Rebalance the tree on remove.
   * @param i the Leaf which must be removed after fixage
   */
  final void fixRemove(MapBase.Leaf i)
  {
    for (MapBase.Leaf j = i.top; j != null; i = j, j = i.top)
    {
      if (j.l == i)
        ++j.len;
      else
        --j.len;

      // change in length
      if (j.len == 0)
        continue;

      // no change in length
      if (j.len == 1 || j.len == -1)
        return;

      if (j.len == -2)
      {
        i = j.l;
        if (i.len > 0)
        {
          drr(i, j);
          j = j.top;
          continue;
        }
        rr(i, j);
        if (i.len != 0)
          return;
        j = i;
        continue;
      }
      if (j.len == 2)
      {
        i = j.r;
        if (i.len < 0)
        {
          drl(i, j);
          j = j.top;
          continue;
        }
        rl(i, j);
        if (i.len != 0)
          return;
        j = i;
        continue;
      }
    }
  }

  MapBase.Leaf find(Object key)
  {
    for (MapBase.Leaf p = root; p != null;)
    {
      int c = compare(key, p.key);
      if (c == 0)
        return p;
      if (c > 0)
        p = p.r;
      else
        p = p.l;
    }
    return null;
  }

  /**
   * finds the next Leaf which either fits the key or is the right next.  
   * @param key
   * @return the matching Leaf or null if none was found.
   */
  MapBase.Leaf findNext(Object key)
  {
    MapBase.Leaf l = null;

    for (MapBase.Leaf p = root; p != null;)
    {
      int c = compare(key, p.key);
      if (c == 0)
        return p;
      if (c > 0) {
        l = p;
        p = p.r;
      }
      else
        p = p.l;
    }

    if (l == null)
      return firstLeaf();
    
    for(;;) {
      l = l.next();
      if (l == null)
        break;
      if (compare(key, l.key) < 0)
        break;
    }
    return l;
  }

  // remove the Leaf from the tree
  final Object unlink(MapBase.Leaf i)
  {
    --count;
    if (count == 0)
    {
      root = null;
      return i.value;
    }
    if (i.l != null && i.r != null)
    {
      // seek replacement
      MapBase.Leaf r = i.len > 0 ? i.next() : i.prev();

      Object o = i.value;
      i.key = r.key;
      i.value = r.value;
      i = r;
      i.value = o;
    }

    // do direct unlink
    fixRemove(i);
    MapBase.Leaf j = i.l == null ? i.r : i.l;
    MapBase.Leaf k = i.top;
    if (k == null)
      root = j;
    else
    {
      if (k.l == i)
      {
        k.l = j;
      } else
      {
        k.r = j;
      }
    }
    if (j != null)
      j.top = k;
    return i.value;
  }

  /**
   * Get the element for the specified value
   * @param key the key for the element to search
   * @return the value if found, either null
   */
  public final Object get(Object key)
  {
    MapBase.Leaf l = find(key);
    if (l == null)
      return null;
    return l.value;
  }

  /**
   * Removes the key holding the specified value from this Map. Useful in MultiMap to remove a distinct key / value entry.
   * This method does nothing if the key is not in the Map.
   * If value is null, the first matching key is removed.
   * @param key the key that needs to be removed.
   * @return the value to which the key had been mapped in this MultiMap, or null if the key did not have a mapping.
   */
  public synchronized Object remove(Object key)
  {
    // search the key
    MapBase.Leaf p = find(key);
    if (p == null)
      return null;
    return unlink(p);
  }

  /**
   * Removes the key holding the specified value from this Map. Useful in MultiMap to remove a distinct key / value entry.
   * This method does nothing if the key is not in the Map.
   * If value is null, the first matching key is removed.
   * @param key the key that needs to be removed.
   * @param value the value at the key that needs to be removed.
   * @return the value to which the key had been mapped in this MultiMap, or null if the key did not have a mapping.
   */
  public abstract Object remove(Object key, Object value);

  protected final int compare(Object a, Object b)
  {
    return comp == null ? ((Comparable) a).compareTo(b) : comp.compare(a, b);
  }

  /**     
   * Returns <tt>true</tt> if this map contains a mapping for the specified     
   * key.     
   *     
   * @param key key whose presence in this map is to be tested.     
   * @return <tt>true</tt> if this map contains a mapping for the specified     
   * key.     
   *      
   * @throws NullPointerException if the key is <tt>null</tt> and this map     
   *            does not not permit <tt>null</tt> keys.     
   */
  public boolean containsKey(Object key)
  {
    return get(key) != null;
  }

  /**
   * Return a Set for this Map's entries.
   * @return a Set for this Map's entries.
   */
  public java.util.Set entrySet()
  {
    return new SubSet(null, null, this);
  }

  /**
   * Return the used Comparator object, if any.
   * @return the used Comparator object, or null.
   */
  public java.util.Comparator comparator()
  {
    return comp;
  }

  /**
   * Return the key of the sorted first entry in this Map.
   * @return the key of the sorted first entry in this Map, or null on empty Map.
   */
  public java.lang.Object firstKey()
  {
    if (root == null)
      return null;
    return firstLeaf().key;
  }

  /**
   * Return the key of the sorted last entry in this Map.
   * @return the key of the sorted last entry in this Map, or null on empty Map.
   */
  public java.lang.Object lastKey()
  {
    if (root == null)
      return null;
    return lastLeaf().key;
  }

  /**
   * Create a sub map view of this map. The parameters specify the start and the not included end point.
   * It is also possible to create a SubMap from this SubMap.
   * @param from first key of the sub map
   * @param to first key <b>behind</b> the last key.
   * @return a new allocated object which is a view to the underlying map.
   */
  public java.util.SortedMap subMap(java.lang.Object from, java.lang.Object to)
  {
    return new SubMap(from, to, this);
  }

  /**
   * Create a sub map view of this map. The parameter specifies the start point, and the sub map ends at the same point as this map.
   * It is also possible to create a SubMap from this SubMap.
   * @param from first key of the sub map
   * @return a new allocated object which is a view to the underlying map.
   */
  public java.util.SortedMap tailMap(java.lang.Object from)
  {
    return subMap(from, null);
  }

  /**
   * Create a sub map view of this map. The sub map starts at the same point as this map, and the parameter specifies the not included end point.
   * It is also possible to create a SubMap from this SubMap.
   * @param to first key <b>behind</b> the last key.
   * @return a new allocated object which is a view to the underlying map.
   */
  public java.util.SortedMap headMap(java.lang.Object to)
  {
    return subMap(null, to);
  }

  /**
   * displays the members with toString().
   * @return a String with all members.
   */
  public String toString()
  {
    if (count > 128)
      return "{... more than 128 items ...}";
    StringBuffer sb = new StringBuffer("{");
    Leaf l = firstLeaf();
    while (l != null)
    {
      sb.append(l.key).append('=').append(l.value);
      l = l.next();
      if (l != null)
        sb.append(", ");
    }
    return sb.append("}").toString();
  }

  private static class SubMap extends java.util.AbstractMap implements java.util.SortedMap {
    private MapBase mb;

    private Object fromKey, toKey;

    SubMap(Object from, Object to, MapBase _mb)
    {
      fromKey = from;
      toKey = to;
      mb = _mb;
    }

    private boolean inRange(Object key)
    {
      if (fromKey != null && mb.compare(fromKey, key) > 0)
        return false;
      if (toKey != null && mb.compare(toKey, key) <= 0)
        return false;
      return true;
    }

    /** (non-Javadoc)
     * @see java.util.SortedMap#comparator()
     */
    public java.util.Comparator comparator()
    {
      return mb.comparator();
    }

    /** (non-Javadoc)
     * @see java.util.SortedMap#firstKey()
     */
    public java.lang.Object firstKey()
    {
      if (fromKey == null)
        return mb.firstKey();
      MapBase.Leaf l = mb.findNext(fromKey);
      if (l == null || !inRange(l.key))
        return null;
      return l.key;
    }

    /** (non-Javadoc)
     * @see java.util.SortedMap#headMap(java.lang.Object)
     */
    public java.util.SortedMap headMap(java.lang.Object to)
    {
      if (!inRange(to))
        throw new IllegalArgumentException("key out of range");
      return new SubMap(fromKey, to, mb);
    }

    /** (non-Javadoc)
     * @see java.util.SortedMap#lastKey()
     */
    public java.lang.Object lastKey()
    {
      if (toKey == null)
        return null;
      MapBase.Leaf l = mb.findNext(toKey);
      if (l == null) {
        l = mb.lastLeaf();
        if (inRange(l.key))
          return l.key;
        return null;
      }
      MapBase.Leaf l2 = l.prev();
      if (l2 == null || !inRange(l2.key))
        return firstKey();
      return l.key;
    }

    /** (non-Javadoc)
     * @see java.util.SortedMap#subMap(java.lang.Object, java.lang.Object)
     */
    public java.util.SortedMap subMap(java.lang.Object from, java.lang.Object to)
    {
      if (!inRange(from) || !inRange(to))
        throw new IllegalArgumentException("key out of range");
      return new SubMap(from, to, mb);
    }

    /** (non-Javadoc)
     * @see java.util.SortedMap#tailMap(java.lang.Object)
     */
    public java.util.SortedMap tailMap(java.lang.Object from)
    {
      if (!inRange(from))
        throw new IllegalArgumentException("key out of range");
      return new SubMap(from, toKey, mb);
    }

    /** (non-Javadoc)
     * @see java.util.Map#entrySet()
     */
    public java.util.Set entrySet()
    {
      return new SubSet(fromKey, toKey, mb);
    }

    /** (non-Javadoc)
     * @see java.util.Map#size()
     */
    public int size()
    {
      return entrySet().size();
    }

    /** (non-Javadoc)
     * @see java.util.Map#isEmpty()
     */
    public boolean isEmpty()
    {
      return firstKey() == null;
    }

    /** (non-Javadoc)
     * @see java.util.Map#containsKey(java.lang.Object)
     */
    public boolean containsKey(Object key)
    {
      return inRange(key) && mb.containsKey(key);
    }

    /** (non-Javadoc)
     * @see java.util.Map#get(java.lang.Object)
     */
    public Object get(Object key)
    {
      if (!inRange(key))
        return null;
      return mb.get(key);
    }

    /** (non-Javadoc)
     * @see java.util.Map#put(java.lang.Object, java.lang.Object)
     */
    public Object put(Object key, Object value)
    {
      if (!inRange(key))
        throw new IllegalArgumentException("key out of range");
      return mb.put(key, value);
    }

    /** (non-Javadoc)
     * @see java.util.Map#remove(java.lang.Object)
     */
    public Object remove(Object key)
    {
      if (!inRange(key))
        throw new IllegalArgumentException("key out of range");
      return mb.remove(key);
    }
  }
}

/******************************************************************************
 * $Log: MapBase.java,v $
 * Revision 1.14  2008/03/15 18:01:05  bebbo
 * @R Changed the license: From now on GPL 3 applies.
 *
 * Revision 1.13  2008/03/13 20:51:39  bebbo
 * @B fixed the lastKey() method
 *
 * Revision 1.12  2006/05/09 08:56:06  bebbo
 * @B fixed Iterator NPE
 *
 * Revision 1.11  2006/03/17 11:38:38  bebbo
 * @B fixed Subset and iterators for Subsets
 *
 * Revision 1.10  2006/02/02 07:50:07  bebbo
 * @I cleanup
 *
 * Revision 1.9  2005/02/13 15:16:32  bebbo
 * @I nothing important: beautified imports, added SerialUID
 *
 * Revision 1.8  2004/12/07 15:22:20  bebbo
 * @N added debug friendly toString()
 *
 * Revision 1.7  2003/09/30 13:17:19  bebbo
 * @B fixed api doc
 *
 * Revision 1.6  2003/07/30 10:12:31  bebbo
 * @R now Serializable
 *
 * Revision 1.5  2003/06/17 14:14:44  bebbo
 * @B bug fix in MapBase.Iter.remove()
 *
 * Revision 1.4  2002/09/09 09:40:36  bebbo
 * @B fixed findNext in MultiMap
 *
 * Revision 1.3  2002/08/20 15:17:07  bebbo
 * @I comments and formatting
 *
 * Revision 1.2  2001/11/04 18:36:12  franke
 * @I explicit type naming: MapBase.Leaf instead of Leaf
 *
 * Revision 1.1  2001/09/15 08:57:37  bebbo
 * @N new
 *
 * Revision 1.2  2000/12/29 17:47:24  bebbo
 * @? dunno
 *
 * Revision 1.1  2000/11/09 15:06:17  bebbo
 * @R repackaged
 *
 *****************************************************************************/
