/******************************************************************************
 * $Source: /export/CVS/java/bb_util/src/de/bb/util/MultiMap.java,v $
 * $Revision: 1.9 $
 * $Date: 2008/03/15 18:01:05 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 *
 * Map implementation which allows duplicate keys. 
 *
 * Copyright (c) by Stefan Bebbo Franke 1999-2008.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *****************************************************************************/

package de.bb.util;

import java.util.*;

/**
 * This class is used to maintain key value pairs, sorted by key.
 * All keys must be usable by the specified Comparator object, or if no Comparator is used, by the Objects compareTo function.
 * <br><b>This implementation allows duplicate keys.</b>
 * <ul><li>If a key is inserted twice or more often, insert places the new entry behind (rightmost) the existing entries.
 * </li><li>If a key is removed, the first (leftmost) entry is removed. 
 * </li><li>To delete an explicit key, value pair, a remove(k,p) function exists.</li></ul>
 * In other words: <br>If only one constant key is used, to insert, get and remove values, it acts like a FIFO.
 */
public class MultiMap extends MapBase
{
  /**
	 * Comment for <code>serialVersionUID</code>
	 */
	private static final long serialVersionUID = 3545511807420020020L;

/**
   * Creates a MultiMap object, using the specified Comparator.
   * @param comp_ a Comparator object.
   */
  public MultiMap(Comparator comp_)
  {
    super(comp_);
  }
  /**
   * Creates a MultiMap object without a Comparator.
   */
  public MultiMap()
  {
  }

  /**
   * Removes the key holding the specified value from this Map. Useful in MultiMap to remove a distinct key / value entry.
   * This method does nothing if the key is not in the Map.
   * If value is null, the first matching key is removed.
   * @param key the key that needs to be removed.
   * @param value the value at the key that needs to be removed.
   * @return the value to which the key had been mapped in this MultiMap, or null if the key did not have a mapping.
   */    
  public Object remove(Object key, Object value)
  {
    for (Leaf i = find(key); i != null; i = i.next())
    {
      if (compare(key, i.key) != 0)
        return null;
      if (i.value.equals(value))
        return unlink(i);
    }
    return null;
  }

  /**
   * Really finds the first entry with the possible duplicate key.
   * @param key a key
   * @return the really first Leaf with that key, or null
   */
  Leaf find(Object key)
  {
    Leaf p = root;
    Leaf f = null;
    for(;;)
    {
      int c = 1, lc = 1;
      while (p!=null) {
        while (p!=null && (c = compare(p.key, key))>=0) {
          f = p;
          p = p.l;
          lc = c;
        }
        if (p==null) break;
        p = p.r;
      }
//      if (f==null || compare(f.key, key)>0)
//      if (f == null || c > 0)
      if (f == null || lc != 0)
      {
        return p;
      }        
        
  // handle special case: left < p, but a right child from p might be equal!
      p = f.l;
      while (p!=null && compare(p.key, key)<0)
        p = p.r;
      if (p==null)
        return f;
      f = null;
    }    
  }
  
  MapBase.Leaf findNext(Object key)
  {
    MapBase.Leaf l = find(key);
    if (l != null)
      return l;
    return super.findNext(key);
  }
  
  /**
   * Insert a given object into the tree using the specified key.
   * @param key the key
   * @param value the inserted value
   * @return the old replaced value or null.
   */
  public Object put(Object key, Object value) {
  // handle empty tree: initialize root and count
    if (root == null) {
      root = new Leaf(key, value);
      count = 1;
      return null;
    }

  // insert into tree
    Leaf  neu, i = root;
    for(;;)
    {
      if (compare(key, i.key)<0) {
        if (i.l == null)  {
          neu = i.l = new Leaf(key, value);
          break;
        }
        i = i.l;
      } else {
        if (i.r == null) {
          neu = i.r = new Leaf(key, value);
          break;
        }
        i = i.r;
      }
    }
    
    ++count;
    neu.top = i;
    fixAdd(neu);
    return null;
  }
}

/******************************************************************************
 * $Log: MultiMap.java,v $
 * Revision 1.9  2008/03/15 18:01:05  bebbo
 * @R Changed the license: From now on GPL 3 applies.
 *
 * Revision 1.8  2005/02/13 15:16:32  bebbo
 * @I nothing important: beautified imports, added SerialUID
 *
 * Revision 1.7  2002/09/09 09:40:36  bebbo
 * @B fixed findNext in MultiMap
 *
 * Revision 1.6  2002/03/30 15:40:13  franke
 * @B fix in find() method.
 * If key was not an element of the map, sometimes a wrong value was returned instead of null
 *
 * Revision 1.5  2001/12/10 16:22:50  bebbo
 * @C completed comments!
 *
 * Revision 1.4  2001/11/04 18:36:52  franke
 * @B value is now compared using equals()
 *
 * Revision 1.3  2001/09/15 08:57:52  bebbo
 * @R complete new design, based on MapBase
 *
 * Revision 1.2  2000/12/29 17:47:24  bebbo
 * @? dunno
 *
 * Revision 1.1  2000/11/09 15:06:17  bebbo
 * @R repackaged
 *
 *****************************************************************************/
 