/******************************************************************************
 * $Source: /export/CVS/java/bb_util/src/de/bb/util/SessionManager.java,v $
 * $Revision: 1.22 $
 * $Date: 2008/03/15 18:01:05 $
 * $Author: bebbo $
 * $Locker:  $
 * $State: Exp $
 *
 * Hold entries with a timeout and support refresh, validation and discarding of entries.  
 *
 * Copyright (c) by Stefan Bebbo Franke 1999-2008.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *****************************************************************************/

package de.bb.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;

/**
 * A SessionManager is used to manage sessions. Sessions are objects which live a distinct time.
 * Each time a session is used or touched, the livetime of that session is renewed.
 * If the life time of a session is over, it is removed from the SessionManager.
 * A session object may implement the <code>Session.Callback</code> interface, and if it does,
 * its callback function is called before removal.
 */
public class SessionManager
{
  private final static boolean DEBUG = false;

  private int maxCount;
  // max count of sessions herein, or -1 for unlimited count
  private long timeout; // default timeout for all 
  private HashMap ht; // key -> value map
  private HashMap rev; // reverse: key -> timeout

  private static Random rand = new Random();
  // add some random to all sessionIds

  /** the thread who watches all sessions */
  private static Killer killer = new Killer();

  /** timeout -> key map */
  static MultiMap map = new MultiMap();

  static {
    killer.start();

    // try to use a secure random
    try
    {
      Class clazz = Class.forName("de.bb.security.SecureRandom");
      rand = (Random) clazz.newInstance();
    } catch (Exception ex)
    {
    }
  }

  /**
   * Set a new random generator for all SessionManagers.
   * @param rnd the new random generator
   */
  public static void setRandom(Random rnd)
  {
    if (rnd != null)
      rand = rnd;
  }

  /**
   * This interface is used by the SessionManager, which maintains objects which live a distinct time.
   * To support the <i>soft</i> removeButAsk() function, an object might implement this interface.
   * The callback function dontRemove will ve called before <i>soft</i> removal.
   */
  public static interface Callback
  {
    /**
     * Is called be the removeButAsk() function.
     * If the implementation returns true, the object is not removed.
     * The implementation may touch the object, to renew the sessions life time.
     * If the session remains untouched, touch() is called by the SessionManager with its default life time.
     * @param key the key which binds the session to the SessionManager.
     * @return true to renew the session life time, false, to allow deletion of the session.
     */
    public boolean dontRemove(Object key);
  }

  /**
   * Waits until the time is come to kill the next session
   */
  private static class Killer extends Thread
  {
    /**
     * create a Killer object which knows it's SessionManager
     */
    Killer()
    {
      super("the bb_util SessionManager thread");
      setDaemon(true);
    }
    /**
     * run function, to handle sleep / awake stuff properly.
     */
    public void run()
    {
      for (;;)
      {
        try
        {
          long diff;
          do
          {
            Object[] o = null;
            synchronized (SessionManager.map)
            {
              Long l = (Long) map.firstKey();
              if (l == null)
              {
                diff = 500000;
                break;
              }
              diff = l.longValue() - System.currentTimeMillis();
              if (diff > 0)
                break;
              o = (Object[]) map.get(l);
            }
            if (DEBUG)
              System.out.println("diff = " + diff);
            SessionManager sm = (SessionManager) o[0];
            sm.removeButAsk(o[1]);
          } while (diff <= 0);

          if (diff > 0)
          {
            synchronized (this)
            {
              wait(diff);
            }
          }
        } catch (InterruptedException ie)
        {
          if (DEBUG)
            System.out.println("Killer - wake up");
        }
      }
    }
  }
  /**
   * This class is used to maintain sessions, that are objects which live a distinct time.
   * Each time a session is used or touched, the livetime of that session is renewed.
   * Constructs a SessionManager object.
   *
   * @param timeoutMilli  an int specifying the timeout in ms
   * @param maxCountOfSessions limit the count of used sessions. 
   */
  public SessionManager(long timeoutMilli, int maxCountOfSessions)
  {
    maxCount = maxCountOfSessions;
    timeout = timeoutMilli;
    ht = new HashMap();
    rev = new HashMap();
  }

  /**
   * This class is used to maintain sessions, that are objects which live a distinct time.
   * Each time a session is used or touched, the livetime of that session is renewed.
   * Constructs a SessionManager object.
   *
   * @param timeoutMilli  an int specifying the timeout in ms
   */
  public SessionManager(long timeoutMilli)
  {
    this(timeoutMilli, -1);
  }
  /**
   * This class is used to maintain sessions, that are objects which live a distinct time.
   * Each time a session is used or touched, the livetime of that session is renewed.
   * Constructs a SessionManager object.
   *
   * @param timeoutMilli  an int specifying the timeout in ms
   * @deprecated use the constructor with long instead!
   */
  public SessionManager(int timeoutMilli)
  {
    this(timeoutMilli, -1);
  }
  /**
   * Clears this SessionManager so that it contains no sessions.
   */
  public void clear()
  {
    synchronized (map)
    {
      ht.clear();
      rev.clear();
    }
  }
  /**
   * Tests if some key maps into the specified value in this SessionManager. 
   * This operation is more expensive than the containsKey method.
   * @param session a sessin to search for
   * @return true if some key maps to the value argument in this SessionManager; false otherwise.
   */
  public boolean contains(Object session)
  {
    return ht.containsValue(session);
  }
  /**
   * Tests whether the specified object is a key in this SessionManager.
   * @param key possible key.
   * @return true if the specified object is a key in this SessionManager; false otherwise.
   */
  public boolean containsKey(Object key)
  {
    return ht.containsKey(key);
  }
  /**
   * Returns an enumeration of the values in this SessionManager. 
   * Use the Enumeration methods on the returned object to fetch the elements sequentially.
   * @return an enumeration of the values in this SessionManager.
   */
  public Iterator elements()
  {
    return ht.values().iterator();
  }
  /**
   * Returns the value to which the specified key is mapped in this SessionManager.
   * @param key a key in the SessionManager. 
   * @return the value to which the key is mapped in this SessionManager; null if the key is not mapped to any value in this SessionManager.
   */
  public Object get(Object key)
  {
    return key == null ? null : ht.get(key);
    /*
        if (key == null)
          return null;
        //    synchronized (ht) {  
        Object o = ht.get(key);
        if (DEBUG)
          System.out.println(
            "get ("
              + ht.size()
              + ") "
              + key.toString()
              + " = "
              + (o == null ? "null" : o.toString()));
        return o;
        //    }
    */
  }
  /**
   * Get current max count of sessions.
   * @return current max count of sessions
   * @see #setMaxCount
   */
  public int getMaxCount()
  {
    return maxCount;
  }
  /**
   * Get current default timout value.
   * @return current default timout value.
   * @see #setTimeout
   */
  public long getTimeout()
  {
    return timeout;
  }
  /**
   * Tests if this SessionManager maps no keys to values.
   * @return true if this SessionManager maps no keys to values; false otherwise.
   */
  public boolean isEmpty()
  {
    return ht.isEmpty();
  }
  /**
   * Returns an enumeration of the keys in this SessionManager.
   * @return an enumeration of the keys in this SessionManager.
   */
  public Iterator keys()
  {
    return ht.keySet().iterator();
  }
  /**
   * Maps the specified key to the specified value in this SessionManager. 
   * If the key is null a new key is generated and returned.
   * The value can NOT be null. 
   * The value can be retrieved by calling the get method with a key that is equal to the original key.
   * @param key the key or null, then a new key is created. 
   * @param value the value. 
   * @return the used key to store the value.
   */
  public Object put(Object key, Object value)
  {
    return put(key, value, timeout);
  }

  /**
   * Maps the specified key to the specified value in this SessionManager, using the SessionManager's timeout.
   * If the key is null a new key is generated and returned.
   * The value can NOT be null. 
   * The value can be retrieved by calling the get method with a key that is equal to the original key.
   * @param key the key or null, then a new key is created. 
   * @param value the value. 
   * @param timeoutMilli a distinct timeout value for the given key
   * @return the used key to store the value.
   * @deprecated
   */
  public Object put(Object key, Object value, int timeoutMilli)
  {
    return put(key, value, (long)timeoutMilli);
  }
  /**
   * Maps the specified key to the specified value in this SessionManager, using the SessionManager's timeout.
   * If the key is null a new key is generated and returned.
   * The value can NOT be null. 
   * The value can be retrieved by calling the get method with a key that is equal to the original key.
   * @param key the key or null, then a new key is created. 
   * @param value the value. 
   * @param timeoutMilli a distinct timeout value for the given key
   * @return the used key to store the value.
   */
  public Object put(Object key, Object value, long timeoutMilli)
  {
    synchronized (map)
    {

      // limit count of elements
      if (maxCount >= 0 && maxCount <= ht.size()) {
        // remove the oldest element
        Object o3[] = (Object[])map.get(map.lastKey());
        remove(o3[1]); 
      }

      if (key == null)
        key = newKey();
      Object oVal = ht.put(key, value);
      if (oVal != null)
      {
        if (DEBUG)
          System.out.println("double put!");
        Object tkey = rev.get(key);
        map.remove(tkey);
      }

      long now = System.currentTimeMillis();

      Object o3[] = new Object[] { this, key, new Long(now)};
      Long tkey = new Long(now + timeoutMilli);

      update(key, tkey, o3);

      if (DEBUG)
        size();
      if (DEBUG)
        System.out.println("put k,v:" + key + "," + value);

      return key;
    }
  }

  /**
   * Method update.
   * @param key
   * @param tkey
   * @param o3
   */
  private void update(Object key, Long tkey, Object o3)
  {
    Long tkeyOrg = tkey;

    while (map.get(tkey) != null)
    {
      tkey = new Long(tkey.longValue() + 1);
    }

    rev.put(key, tkey);
    map.put(tkey, o3);

    if (tkey == tkeyOrg)
    {
      // eine neue kleinste Weckzeit?      
      if (tkey.equals(map.firstKey()))
      {
        synchronized (killer)
        {
          killer.notify();
        }
      }
    }
  }

  /**
   * Returns the creation time for the given key.
   * This is the system time in milli seconds when the first put with that key occured.
   * @param key the key
   * @return creation time or -1 on invalid key.
   */
  public long getCreationMillis(Object key)
  {
    Object tkey = rev.get(key);
    if (tkey != null)
    {
      Object[] o3 = (Object[]) map.get(tkey);
      if (o3 != null)
      {
        Long l = (Long) o3[2];
        return l.longValue();
      }
    }
    return -1;
  }

  /**
   * Returns the death time for the given key.
   * This is the system time in milli seconds when the object gets invalid.
   * If the stored object implements the Callback interface, thie implementation
   * of the Callback decides whether its lifetime gets renewed,
   * @param key the key
   * @return death time or -1 on invalid key.
   */
  public long getDeathMillis(Object key)
  {
    Long tkey = (Long) rev.get(key);
    if (tkey != null)
    {
      return tkey.longValue();
    }
    return -1;
  }

  /**
   * Removes the key (and its corresponding value) from this SessionManager. 
   * This method does nothing if the key is not in the SessionManager.
   * If value is an instance of Callback, its dontRemove() method is called, with key as param, then the removal depends on the result of that function.
   * @param key the key that needs to be removed.
   * @return the value to which the key had been mapped in this SessionManager, or null if the key did not have a mapping.
   */
  public Object removeButAsk(Object key)
  {
    Object val, tkey;
    synchronized (map)
    {
      val = ht.get(key);
      if (val == null)
      {
        if (DEBUG)
          System.out.println("NOT FOUND");
        return null;
      }
      tkey = rev.get(key);
      if (val instanceof SessionManager.Callback)
      {
        if (DEBUG)
          System.out.println("notify k,v:" + key + "," + val);

        try
        {
          if (((SessionManager.Callback) val).dontRemove(key))
          {
            Object tkey2 = rev.get(key);
            if (tkey == null || tkey2 == null || !tkey.equals(tkey2))
              return null;
            if (DEBUG)
              System.out.println(
                "K,V,T: "
                  + key
                  + ","
                  + val
                  + ","
                  + tkey
                  + ": not touched -> touch it");
            touch(key);
            return null;
          }
        } catch (Exception ex)
        {
        }
      }
    }
    return remove(key);
  }

  /**
   * Removes the key (and its corresponding value) from this SessionManager. 
   * This method does nothing if the key is not in the SessionManager.
   * @param key the key that needs to be removed.
   * @return the value to which the key had been mapped in this SessionManager, or null if the key did not have a mapping.
   */
  public Object remove(Object key)
  {
    synchronized (map)
    {
      Object val = ht.get(key);
      if (val == null)
        return null;

      if (DEBUG)
        System.out.println("remove k,v:" + key + "," + val);
      Object tkey = rev.get(key);

      rev.remove(key);
      map.remove(tkey);
      Object o = ht.remove(key);
      if (DEBUG)
        size();
      return o;
    }
  }

  /**
   * Set a new maximum size.
   * The new size only affects future invokations of put(...)!
   * @param max the new max count of sessions
   * @see #getMaxCount
   */
  public void setMaxCount(int max)
  {
    maxCount = max;
  }
  /**
   * Set a new default timout value.
   * The new timout only affects future invokations of put(...) or touch(...)!
   * @param timeoutMilli the new default timout value
   * @see #getTimeout
   */
  public void setTimeout(long timeoutMilli)
  {
    timeout = timeoutMilli;
  }
  /**
   * Returns the number of keys in this SessionManager.
   * @return the number of keys in this SessionManager.
   */
  public int size()
  {
    if (DEBUG)
      if (ht.size() != map.size())
        System.out.println(
          "size ht,map,rev: "
            + ht.size()
            + ","
            + map.size()
            + ","
            + rev.size());
    return ht.size();
  }
  /**
   * Sets the new timeout for the given key with the default timeout.
   * @param key the key
   * @return the key or null if the key was not found in the SessionManager.
   */
  public Object touch(Object key)
  {
    return touch(key, timeout);
  }
  /**
   * Sets the new timeout for the given key.
   * @param key the key
   * @param timeoutMilli the new timeout to set
   * @return the key or null if the key was not found in the SessionManager.
   */
  public Object touch(Object key, long timeoutMilli)
  {
    if (key == null)
      return null;
    synchronized (map)
    {
      Object val = ht.get(key);
      if (null == val)
        return null;

      Long tkey = (Long) rev.get(key);
      Object o3 = map.remove(tkey);

      Long tkeyNew = new Long(System.currentTimeMillis() + timeoutMilli);

      update(key, tkeyNew, o3);

      return key;
    }
  }

  private static String zeros = "0000000000000000";
  /**
   * Create a new session key.
   * @return a new created String, containing a
   */
  public static String newKey()
  {
    String lh = Long.toHexString(rand.nextLong());
    return Long.toHexString(System.currentTimeMillis())
      + zeros.substring(lh.length())
      + lh;
  }

}

/******************************************************************************
 * $Log: SessionManager.java,v $
 * Revision 1.22  2008/03/15 18:01:05  bebbo
 * @R Changed the license: From now on GPL 3 applies.
 *
 * Revision 1.21  2007/04/20 08:06:57  bebbo
 * @R if the limit of elements is reached now the oldest elements are dropped.
 *
 * Revision 1.20  2004/12/01 12:39:31  bebbo
 * @I replaced Hashtable with HashMap
 *
 * Revision 1.19  2004/04/14 13:35:26  bebbo
 * @B added old CT and put methode with int, but marked as deprecated
 *
 * Revision 1.18  2004/03/14 18:02:29  bebbo
 * @R changed timeout to long value
 *
 * Revision 1.17  2003/02/17 14:23:08  bebbo
 * @I changed access rights to remove synthetic accessor methods
 *
 * Revision 1.16  2002/12/10 13:06:57  bebbo
 * @B fixed shorting of killer thread wait time - was never done
 *
 * Revision 1.15  2002/11/27 14:04:43  bebbo
 * @I reformat
 *
 * Revision 1.14  2002/11/13 16:32:55  bebbo
 * @B get(null) now returns only null (no more Exception)
 *
 * Revision 1.13  2002/11/13 16:07:20  bebbo
 * @B fixed erraneous synchronized statements - could cause deadlocks
 *
 * Revision 1.12  2002/10/24 17:26:38  bebbo
 * @N added new Method getCreationMillis
 * @N added new Method getDeathMillis
 *
 * Revision 1.11  2002/09/03 11:50:40  bebbo
 * @B fixed usage of only one thread
 *
 * Revision 1.10  2002/08/21 14:49:27  bebbo
 * @I all SessionManager instances are using the same wakeup thread
 *
 * Revision 1.9  2002/08/20 15:17:57  bebbo
 * @I formatting
 * @N new function destroy
 *
 * Revision 1.8  2001/12/10 16:22:50  bebbo
 * @C completed comments!
 *
 * Revision 1.7  2001/11/04 18:37:28  franke
 * @I explicit type naming for inner classes
 *
 * Revision 1.6  2001/09/15 08:56:56  bebbo
 * @C added comments
 *
 * Revision 1.5  2001/08/03 13:41:29  bebbo
 * @B removed erraneous synchronized statements
 *
 * Revision 1.4  2001/03/29 19:57:03  bebbo
 * @B fixed Nullpointer exceptions when key is null
 *
 * Revision 1.3  2000/12/30 09:04:05  bebbo
 * @I now running as a daemon
 *
 * Revision 1.2  2000/12/29 17:47:24  bebbo
 * @? dunno
 *
 * Revision 1.1  2000/11/10 18:12:18  bebbo
 * @R repackaged
 *
 *****************************************************************************/